@@ -162,6 +162,7 @@ static int parse_input_node(struct openconnect_info *vpninfo, struct oc_auth_for
 			vpn_progress(vpninfo, PRG_DEBUG,
 				     _("Discarding duplicate option '%s'\n"),
 				     opt->name);
+			free_opt(opt);
 			goto out;
 		}
 		p = &(*p)->next;
@@ -197,8 +198,10 @@ static int parse_select_node(struct openconnect_info *vpninfo, struct oc_auth_fo
 			continue;
 
 		choice = calloc(1, sizeof(*choice));
-		if (!choice)
+		if (!choice) {
+			free_opt((void *)opt);
 			return -ENOMEM;
+		}
 
 		xmlnode_get_prop(node, "name", &choice->name);
 		choice->label = (char *)xmlNodeGetContent(child);
@@ -310,7 +313,11 @@ static int check_cookie_success(struct openconnect_info *vpninfo)
 		buf_append(buf, "Cookie=%s\n", dspreauth);
 		if (buf_error(buf))
 			return buf_free(buf);
-		send(vpninfo->tncc_fd, buf->data, buf->pos, 0);
+		if (send(vpninfo->tncc_fd, buf->data, buf->pos, 0) < 0) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to send cookie to TNCC\n"));
+			/* Continue anyway */
+		}
 		buf_truncate(buf);
 	}
 
@@ -601,8 +608,10 @@ int oncp_obtain_cookie(struct openconnect_info *vpninfo)
 	int try_tncc = !!vpninfo->csd_wrapper;
 
 	resp_buf = buf_alloc();
-	if (buf_error(resp_buf))
-		return -ENOMEM;
+	if (buf_error(resp_buf)) {
+		ret = buf_error(resp_buf);
+		goto out;
+	}
 
 	while (1) {
 		char *form_buf = NULL;
