@@ -156,11 +156,13 @@ enum {
 	OPT_COOKIE_ON_STDIN,
 	OPT_CSD_USER,
 	OPT_CSD_WRAPPER,
+	OPT_CIPHERSUITES,
 	OPT_DISABLE_IPV6,
 	OPT_DTLS_CIPHERS,
 	OPT_DTLS12_CIPHERS,
 	OPT_DUMP_HTTP,
 	OPT_FORCE_DPD,
+	OPT_FORCE_TROJAN,
 	OPT_GNUTLS_DEBUG,
 	OPT_JUNIPER,
 	OPT_KEY_PASSWORD_FROM_FSID,
@@ -266,6 +268,7 @@ static const struct option long_options[] = {
 	OPTION("no-http-keepalive", 0, OPT_NO_HTTP_KEEPALIVE),
 	OPTION("no-cert-check", 0, OPT_NO_CERT_CHECK),
 	OPTION("force-dpd", 1, OPT_FORCE_DPD),
+	OPTION("force-trojan", 1, OPT_FORCE_TROJAN),
 	OPTION("non-inter", 0, OPT_NON_INTER),
 	OPTION("dtls-local-port", 1, OPT_DTLS_LOCAL_PORT),
 	OPTION("token-mode", 1, OPT_TOKEN_MODE),
@@ -278,6 +281,9 @@ static const struct option long_options[] = {
 	OPTION("form-entry", 1, 'F'),
 #ifdef OPENCONNECT_GNUTLS
 	OPTION("gnutls-debug", 1, OPT_GNUTLS_DEBUG),
+	OPTION("gnutls-priority", 1, OPT_CIPHERSUITES),
+#elif defined(OPENCONNECT_OPENSSL)
+	OPTION("openssl-ciphers", 1, OPT_CIPHERSUITES),
 #endif
 	OPTION(NULL, 0, 0)
 };
@@ -406,8 +412,13 @@ static void read_stdin(char **string, int hidden, int allow_fail)
 		nr_read = wcslen(wbuf);
 	}
 	if (nr_read >= 2 && wbuf[nr_read - 1] == 10 && wbuf[nr_read - 2] == 13) {
+		/* remove trailing "\r\n" */
 		wbuf[nr_read - 2] = 0;
 		nr_read -= 2;
+	} else if (nr_read >= 1 && wbuf[nr_read - 1] == 10) {
+		/* remove trailing "\n" */
+		wbuf[nr_read - 1] = 0;
+		nr_read -= 1;
 	}
 
 	nr_read = WideCharToMultiByte(CP_UTF8, 0, wbuf, -1, NULL, 0, NULL, NULL);
@@ -589,11 +600,7 @@ static void print_build_opts(void)
 {
 	const char *comma = ", ", *sep = comma + 1;
 
-#if defined(OPENCONNECT_OPENSSL)
-	printf(_("Using OpenSSL. Features present:"));
-#elif defined(OPENCONNECT_GNUTLS)
-	printf(_("Using GnuTLS. Features present:"));
-#endif
+	printf(_("Using %s. Features present:"), openconnect_get_tls_library_version());
 
 	if (openconnect_has_tss_blob_support()) {
 		printf("%sTPM", sep);
@@ -681,7 +688,7 @@ static void print_supported_protocols_usage(void)
 static const char default_vpncscript[] = DEFAULT_VPNCSCRIPT;
 static void read_stdin(char **string, int hidden, int allow_fail)
 {
-	char *c, *buf = malloc(1025);
+	char *c, *got, *buf = malloc(1025);
 	int fd = fileno(stdin);
 	struct termios t;
 
@@ -696,7 +703,7 @@ static void read_stdin(char **string, int hidden, int allow_fail)
 		tcsetattr(fd, TCSANOW, &t);
 	}
 
-	buf = fgets(buf, 1025, stdin);
+	got = fgets(buf, 1025, stdin);
 
 	if (hidden) {
 		t.c_lflag |= ECHO;
@@ -704,7 +711,7 @@ static void read_stdin(char **string, int hidden, int allow_fail)
 		fprintf(stderr, "\n");
 	}
 
-	if (!buf) {
+	if (!got) {
 		if (allow_fail) {
 			*string = NULL;
 			free(buf);
@@ -812,8 +819,8 @@ static void usage(void)
 	printf("  -g, --usergroup=GROUP           %s\n", _("Set login usergroup"));
 	printf("  -p, --key-password=PASS         %s\n", _("Set key passphrase or TPM SRK PIN"));
 	printf("      --key-password-from-fsid    %s\n", _("Key passphrase is fsid of file system"));
-	printf("      --token-mode=MODE           %s\n", _("Software token type: rsa, totp or hotp"));
-	printf("      --token-secret=STRING       %s\n", _("Software token secret"));
+	printf("      --token-mode=MODE           %s\n", _("Software token type: rsa, totp, hotp or oidc"));
+	printf("      --token-secret=STRING       %s\n", _("Software token secret or oidc token"));
 #ifndef HAVE_LIBSTOKEN
 	printf("                                  %s\n", _("(NOTE: libstoken (RSA SecurID) disabled in this build)"));
 #endif
@@ -837,7 +844,7 @@ static void usage(void)
 #endif
 	printf("      --reconnect-timeout         %s\n", _("Connection retry timeout in seconds"));
 	printf("      --resolve=HOST:IP           %s\n", _("Use IP when connecting to HOST"));
-	printf("      --passtos                   %s\n", _("copy TOS / TCLASS when using DTLS"));
+	printf("      --passtos                   %s\n", _("Copy TOS / TCLASS field into DTLS and ESP packets"));
 	printf("      --dtls-local-port=PORT      %s\n", _("Set local port for DTLS and ESP datagrams"));
 
 	printf("\n%s:\n", _("Authentication (two-phase)"));
@@ -878,7 +885,7 @@ static void usage(void)
 	printf("      --base-mtu=MTU              %s\n", _("Indicate path MTU to/from server"));
 	printf("  -d, --deflate                   %s\n", _("Enable stateful compression (default is stateless only)"));
 	printf("  -D, --no-deflate                %s\n", _("Disable all compression"));
-	printf("      --force-dpd=INTERVAL        %s\n", _("Set minimum Dead Peer Detection interval"));
+	printf("      --force-dpd=INTERVAL        %s\n", _("Set minimum Dead Peer Detection interval (in seconds)"));
 	printf("      --pfs                       %s\n", _("Require perfect forward secrecy"));
 	printf("      --no-dtls                   %s\n", _("Disable DTLS and ESP"));
 	printf("      --dtls-ciphers=LIST         %s\n", _("OpenSSL ciphers to support for DTLS"));
@@ -895,6 +902,7 @@ static void usage(void)
 	printf("\n%s:\n", _("Trojan binary (CSD) execution"));
 	printf("      --csd-user=USER             %s\n", _("Drop privileges during trojan execution"));
 	printf("      --csd-wrapper=SCRIPT        %s\n", _("Run SCRIPT instead of trojan binary"));
+	printf("      --force-trojan=INTERVAL     %s\n", _("Set minimum interval for rerunning trojan (in seconds)"));
 #endif
 
 	printf("\n%s:\n", _("Server bugs"));
@@ -1447,6 +1455,9 @@ int main(int argc, char **argv)
 		case OPT_FORCE_DPD:
 			openconnect_set_dpd(vpninfo, atoi(config_arg));
 			break;
+		case OPT_FORCE_TROJAN:
+			openconnect_set_trojan_interval(vpninfo, atoi(config_arg));
+			break;
 		case OPT_DTLS_LOCAL_PORT:
 			vpninfo->dtls_local_port = atoi(config_arg);
 			break;
@@ -1459,6 +1470,8 @@ int main(int argc, char **argv)
 				token_mode = OC_TOKEN_MODE_HOTP;
 			} else if (strcasecmp(config_arg, "yubioath") == 0) {
 				token_mode = OC_TOKEN_MODE_YUBIOATH;
+			} else if (strcasecmp(config_arg, "oidc") == 0) {
+				token_mode = OC_TOKEN_MODE_OIDC;
 			} else {
 				fprintf(stderr, _("Invalid software token mode \"%s\"\n"),
 					config_arg);
@@ -1494,6 +1507,21 @@ int main(int argc, char **argv)
 			gnutls_global_set_log_function(oc_gnutls_log_func);
 			break;
 #endif
+		case OPT_CIPHERSUITES:
+			fprintf(stderr,
+			        _("WARNING: You specified %s. This should not be\n"
+			          "         necessary; please report cases where a priority string\n"
+			          "         override is necessary to connect to a server\n"
+			          "         to <openconnect-devel@lists.infradead.org>.\n"),
+#ifdef OPENCONNECT_GNUTLS
+				  "--gnutls-priority"
+#elif defined(OPENCONNECT_OPENSSL)
+				  "--openssl-ciphers"
+#endif
+				  );
+
+			strncpy(vpninfo->ciphersuite_config, config_arg, sizeof(vpninfo->ciphersuite_config) - 1);
+			break;
 		default:
 			usage();
 		}
@@ -2123,7 +2151,7 @@ static int lock_token(void *tokdata)
 	int err;
 
 	/* FIXME: Actually lock the file */
-	err = read_file_into_string(vpninfo, token_filename, &file_token);
+	err = openconnect_read_file(vpninfo, token_filename, &file_token);
 	if (err < 0)
 	    return err;
 
@@ -2170,13 +2198,14 @@ static void init_token(struct openconnect_info *vpninfo,
 	int ret;
 	char *file_token = NULL;
 
-	if (token_str) {
+	if (token_str && (token_mode == OC_TOKEN_MODE_TOTP ||
+			  token_mode == OC_TOKEN_MODE_HOTP)) {
 		switch(token_str[0]) {
 		case '@':
 			token_str++;
 			/* fall through... */
 		case '/':
-			if (read_file_into_string(vpninfo, token_str,
+			if (openconnect_read_file(vpninfo, token_str,
 						  &file_token) < 0)
 				exit(1);
 			break;
@@ -2203,7 +2232,10 @@ static void init_token(struct openconnect_info *vpninfo,
 			fprintf(stderr, _("Soft token string is invalid\n"));
 			exit(1);
 		case -ENOENT:
-			fprintf(stderr, _("Can't open ~/.stokenrc file\n"));
+			if (token_str)
+				fprintf(stderr, _("Can't open stoken file\n"));
+			else
+				fprintf(stderr, _("Can't open ~/.stokenrc file\n"));
 			exit(1);
 		case -EOPNOTSUPP:
 			fprintf(stderr, _("OpenConnect was not built with libstoken support\n"));
@@ -2248,6 +2280,19 @@ static void init_token(struct openconnect_info *vpninfo,
 			exit(1);
 		}
 
+	case OC_TOKEN_MODE_OIDC:
+		switch (ret) {
+		case 0:
+			return;
+		case -ENOENT:
+			fprintf(stderr, _("Can't open oidc file\n"));
+			exit(1);
+		default:
+			fprintf(stderr, _("General failure in oidc token\n"));
+			exit(1);
+		}
+
+		break;
 	case OC_TOKEN_MODE_NONE:
 		/* No-op */
 		break;
