@@ -82,11 +82,14 @@ static int cancellable_connect(struct openconnect_info *vpninfo, int sockfd,
 	int maxfd = sockfd;
 	int err;
 
-	set_sock_nonblock(sockfd);
+	if (set_sock_nonblock(sockfd))
+		goto sockerr;
+
 	if (vpninfo->protect_socket)
 		vpninfo->protect_socket(vpninfo->cbdata, sockfd);
 
 	if (connect(sockfd, addr, addrlen) < 0 && !connect_pending()) {
+	sockerr:
 #ifdef _WIN32
 		return WSAGetLastError();
 #else
@@ -103,7 +106,12 @@ static int cancellable_connect(struct openconnect_info *vpninfo, int sockfd,
 		FD_SET(sockfd, &ex_set);
 #endif
 		cmd_fd_set(vpninfo, &rd_set, &maxfd);
-		select(maxfd + 1, &rd_set, &wr_set, &ex_set, NULL);
+		if (select(maxfd + 1, &rd_set, &wr_set, &ex_set, NULL) < 0 &&
+		    errno != EINTR) {
+			vpn_perror(vpninfo, _("Failed select() for socket connect"));
+			return -EIO;
+		}
+
 		if (is_cancel_pending(vpninfo, &rd_set)) {
 			vpn_progress(vpninfo, PRG_ERR, _("Socket connect cancelled\n"));
 			return -EINTR;
@@ -130,7 +138,7 @@ static int cancellable_connect(struct openconnect_info *vpninfo, int sockfd,
 #else
 	err = -errno;
 	if (err == -ENOTCONN) {
-		int ch;
+		char ch;
 
 		if (read(sockfd, &ch, 1) < 0)
 			err = -errno;
@@ -374,6 +382,7 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 						     _("Failed to allocate sockaddr storage\n"));
 					closesocket(ssl_sock);
 					ssl_sock = -ENOMEM;
+					freeaddrinfo(result);
 					goto out;
 				}
 				vpninfo->peer_addrlen = rp->ai_addrlen;
@@ -831,7 +840,11 @@ void poll_cmd_fd(struct openconnect_info *vpninfo, int timeout)
 
 		FD_ZERO(&rd_set);
 		cmd_fd_set(vpninfo, &rd_set, &maxfd);
-		select(maxfd + 1, &rd_set, NULL, NULL, &tv);
+		if (select(maxfd + 1, &rd_set, NULL, NULL, &tv) < 0 &&
+		    errno != EINTR) {
+			vpn_perror(vpninfo, _("Failed select() for command socket"));
+		}
+
 		check_cmd_fd(vpninfo, &rd_set);
 	}
 }
@@ -905,6 +918,66 @@ FILE *openconnect_fopen_utf8(struct openconnect_info *vpninfo, const char *fname
 	return fdopen(fd, mode);
 }
 
+ssize_t openconnect_read_file(struct openconnect_info *vpninfo, const char *fname,
+			      char **ptr)
+{
+	int fd, len;
+	struct stat st;
+	char *buf;
+
+	fd = openconnect_open_utf8(vpninfo, fname, O_RDONLY|O_BINARY);
+	if (fd < 0) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to open %s: %s\n"),
+			     fname, strerror(errno));
+		return -ENOENT;
+	}
+
+	if (fstat(fd, &st)) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to fstat() %s: %s\n"),
+			     fname, strerror(errno));
+		close(fd);
+		return -EIO;
+	}
+
+	if (st.st_size == 0) {
+		vpn_progress(vpninfo, PRG_INFO, _("File %s is empty\n"),
+			     vpninfo->xmlconfig);
+		close(fd);
+		return -ENOENT;
+	}
+	if (st.st_size >= INT_MAX || st.st_size < 0) {
+		vpn_progress(vpninfo, PRG_INFO, _("File %s has suspicious size %zd\n"),
+			     vpninfo->xmlconfig, (ssize_t)st.st_size);
+		close(fd);
+		return -EIO;
+	}
+	len = st.st_size;
+	buf = malloc(len + 1);
+	if (!buf) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to allocate %d bytes for %s\n"),
+			     len + 1, fname);
+		close(fd);
+		return -ENOMEM;
+	}
+
+	if (read(fd, buf, len) != len) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to read %s: %s\n"),
+			     fname, strerror(errno));
+		free(buf);
+		close(fd);
+		return -EIO;
+	}
+
+	buf[len] = 0;
+	close(fd);
+	*ptr = buf;
+	return len;
+}
+
 int udp_sockaddr(struct openconnect_info *vpninfo, int port)
 {
 	free(vpninfo->dtls_addr);
@@ -954,7 +1027,9 @@ int udp_connect(struct openconnect_info *vpninfo)
 		vpninfo->protect_socket(vpninfo->cbdata, fd);
 
 	sndbuf = vpninfo->ip_info.mtu * 2;
-	setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (void *)&sndbuf, sizeof(sndbuf));
+	if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (void *)&sndbuf, sizeof(sndbuf)) < 0) {
+		vpn_perror(vpninfo, "Set UDP socket send buffer");
+	}
 
 	if (vpninfo->dtls_local_port) {
 		union {
@@ -999,7 +1074,11 @@ int udp_connect(struct openconnect_info *vpninfo)
 	}
 
 	set_fd_cloexec(fd);
-	set_sock_nonblock(fd);
+	if (set_sock_nonblock(fd)) {
+		vpn_perror(vpninfo, _("Make UDP socket non-blocking"));
+		closesocket(fd);
+		return -EIO;
+	}
 
 	return fd;
 }
@@ -1102,7 +1181,12 @@ int cancellable_send(struct openconnect_info *vpninfo, int fd,
 		FD_SET(fd, &wr_set);
 		cmd_fd_set(vpninfo, &rd_set, &maxfd);
 
-		select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
+		if (select(maxfd + 1, &rd_set, &wr_set, NULL, NULL) < 0 &&
+		    errno != EINTR) {
+			vpn_perror(vpninfo, _("Failed select() for socket send"));
+			return -EIO;
+		}
+
 		if (is_cancel_pending(vpninfo, &rd_set))
 			return -EINTR;
 
@@ -1137,7 +1221,12 @@ int cancellable_recv(struct openconnect_info *vpninfo, int fd,
 		FD_SET(fd, &rd_set);
 		cmd_fd_set(vpninfo, &rd_set, &maxfd);
 
-		select(maxfd + 1, &rd_set, NULL, NULL, NULL);
+		if (select(maxfd + 1, &rd_set, NULL, NULL, NULL) < 0 &&
+		    errno != EINTR) {
+			vpn_perror(vpninfo, _("Failed select() for socket recv"));
+			return -EIO;
+		}
+
 		if (is_cancel_pending(vpninfo, &rd_set))
 			return -EINTR;
 
